from time import time as _time
from constance import settings, utils


class Config(object):
    """
    The global config wrapper that handles the backend.
    """
    def __init__(self):
        super(Config, self).__setattr__('_backend', utils.import_module_attr(settings.BACKEND)())

    def __getattr__(self, key):
        try:
            opts = settings.CONFIG[key]
        except KeyError:
            raise AttributeError(key)

        result = self._backend.get(key)
        if result is None:
            default = opts['default']
            if not settings.READONLY:
                setattr(self, key, default)
            return default
        return result

    def __setattr__(self, key, value):
        if key not in settings.CONFIG:
            raise AttributeError(key)
        self._backend.set(key, value)

    def __dir__(self):
        return settings.CONFIG.keys()


class CachedConfig(Config):
    """
    Cached version of config

    Cached the resulting config value for CACHE_TIMEOUT seconds.
    """

    def __init__(self):
        super(CachedConfig, self).__init__()
        super(Config, self).__setattr__('_cache', {})

    def __getattr__(self, key):
        try:
            expire, value = self._cache[key]
            if expire > _time():
                return value
        except KeyError:
            pass

        # if key does not exist in _cache or is expired, retreive it from the backend
        value = super(CachedConfig, self).__getattr__(key)
        expire = _time() + settings.CACHE_TIMEOUT
        self._cache[key] = (expire, value)
        return value

    def __setattr__(self, key, value):
        super(CachedConfig, self).__setattr__(key, value)
        expire = _time() + settings.CACHE_TIMEOUT
        self._cache[key] = (expire, value)


class CachedAllConfig(Config):
    """
    Cached version of config that caches all config key, value pairs at once

    Cached all the config values for CACHE_TIMEOUT seconds.
    This has the benefit of being able to do mget() to retreiving the values
    lowering the overhead of retreiving the values. Drawback is that if only a
    few keys are used it will actually cause an additional overhead retreiving
    values which will not be accessed most of the time within the CACHE_TIMEOUT
    period.
    """

    def __init__(self):
        super(CachedAllConfig, self).__init__()
        super(Config, self).__setattr__('_cache', {})
        super(Config, self).__setattr__('_expired', 0)

    def refresh_cache(self):
        new_cache = dict((k, v['default']) for k, v in settings.CONFIG.iteritems())
        new_cache.update(dict(self._backend.mget(settings.CONFIG.keys())))
        super(Config, self).__setattr__('_cache', new_cache)
        super(Config, self).__setattr__('_expired', _time() + settings.CACHE_TIMEOUT)

    def __getattr__(self, key):
        if self._expired < _time():
            self.refresh_cache()

        try:
            return self._cache[key]
        except KeyError:
            pass

        # if key does not exist in _cache, retreive it from the backend
        value = super(CachedAllConfig, self).__getattr__(key)
        self._cache[key] = value
        return value

    def __setattr__(self, key, value):
        super(CachedAllConfig, self).__setattr__(key, value)
        self._cache[key] = value



import six
from datetime import datetime, date, time
from decimal import Decimal

from django import forms
from django.contrib import admin, messages
from django.contrib.admin import widgets
from django.contrib.admin.options import csrf_protect_m
from django.core.exceptions import PermissionDenied
from django.forms import fields
from django.http import HttpResponseRedirect
from django.shortcuts import render_to_response
from django.template.context import RequestContext
from django.utils.formats import localize
from django.utils.translation import ugettext as _

try:
    from django.conf.urls import patterns, url
except ImportError:  # Django < 1.4
    from django.conf.urls.defaults import patterns, url


from constance import config, settings


NUMERIC_WIDGET = forms.TextInput(attrs={'size': 10})

INTEGER_LIKE = (fields.IntegerField, {'widget': NUMERIC_WIDGET})
STRING_LIKE = (fields.CharField, {
    'widget': forms.Textarea(attrs={'rows': 3}),
    #'required': False,
})

FIELDS = {
    bool: (fields.BooleanField, {'required': False}),
    int: INTEGER_LIKE,
    Decimal: (fields.DecimalField, {'widget': NUMERIC_WIDGET}),
    str: STRING_LIKE,
    datetime: (fields.DateTimeField, {'widget': widgets.AdminSplitDateTime}),
    date: (fields.DateField, {'widget': widgets.AdminDateWidget}),
    time: (fields.TimeField, {'widget': widgets.AdminTimeWidget}),
    float: (fields.FloatField, {'widget': NUMERIC_WIDGET}),
}

if six.PY3:
    unicode = str
    long = int

if not six.PY3:
    FIELDS.update({
        long: INTEGER_LIKE,
        unicode: STRING_LIKE,
    })


def false_func(*args, **kwargs):
    return False


class ConstanceForm(forms.Form):
    def __init__(self, *args, **kwargs):
        super(ConstanceForm, self).__init__(*args, **kwargs)
        for name, opts in six.iteritems(settings.CONFIG):
            _type = type(opts['default'])
            field_class, kwargs = FIELDS[_type]
            if _type in (str, unicode) and opts.get('required', True) is False:
                kwargs['required'] = False
            self.fields[name] = field_class(label=name, **kwargs)
            if settings.READONLY:
                self.fields[name].widget.attrs['disabled'] = 'disabled'

    def save(self):
        for name in self.cleaned_data:
            setattr(config, name, self.cleaned_data[name])


class ConstanceAdmin(admin.ModelAdmin):

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.module_name
        return patterns(
            '',

            url(r'^$',
                self.admin_site.admin_view(self.changelist_view),
                name='%s_%s_changelist' % info),
            url(r'^$',
                self.admin_site.admin_view(self.changelist_view),
                name='%s_%s_add' % info),
        )

    @csrf_protect_m
    def changelist_view(self, request, extra_context=None):
        # First load a mapping between config name and default value
        if not self.has_change_permission(request, None):
            raise PermissionDenied
        default_initial = ((name, opts['default'])
                           for name, opts in six.iteritems(settings.CONFIG))
        # Then update the mapping with actually values from the backend
        initial = dict(default_initial, **dict(config._backend.mget(settings.CONFIG.keys())))
        form = ConstanceForm(initial=initial)
        if request.method == 'POST':
            form = ConstanceForm(request.POST)
            if form.is_valid():
                form.save()
                # In django 1.5 this can be replaced with self.message_user
                messages.add_message(
                    request,
                    messages.SUCCESS,
                    _('Live settings updated successfully.'),
                )
                return HttpResponseRedirect('.')
        context = {
            'config': [],
            'title': _('Constance config'),
            'app_label': 'constance',
            'opts': Config._meta,
            'form': form,
            'media': self.media + form.media,
            'readonly': settings.READONLY,
        }
        for name, opts in six.iteritems(settings.CONFIG):
            default = opts['default']
            help_text = opts['help_text']
            # First try to load the value from the actual backend
            value = initial.get(name)
            # Then if the returned value is None, get the default
            if value is None:
                value = getattr(config, name)
            context['config'].append({
                'name': name,
                'default': localize(default),
                'help_text': _(help_text),
                'value': localize(value),
                'modified': value != default,
                'form_field': form[name],
            })
        context_instance = RequestContext(request,
                                          current_app=self.admin_site.name)
        template_name = 'admin/constance/change_list.html'
        return render_to_response(template_name, context,
                                  context_instance=context_instance)

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, request, obj=None):
        if settings.SUPERUSER_ONLY:
            return request.user.is_superuser
        return super(ConstanceAdmin, self).has_change_permission(request, obj)


class Config(object):
    class Meta(object):
        app_label = 'constance'
        object_name = 'Config'
        model_name = module_name = 'config'
        verbose_name_plural = _('config')
        get_ordered_objects = false_func
        abstract = False
        swapped = False

        def get_change_permission(self):
            return 'change_%s' % self.model_name

    _meta = Meta()


admin.site.register([Config], ConstanceAdmin)



import os
from constance.utils import import_module_attr, parse_config

settings = import_module_attr(
    os.getenv('CONSTANCE_SETTINGS_MODULE', 'django.conf.settings')
)

READONLY = getattr(settings, 'CONSTANCE_READONLY', False)

REDIS_PREFIX = getattr(settings, 'CONSTANCE_REDIS_PREFIX', getattr(settings, 'CONSTANCE_PREFIX', 'constance:'))

BACKEND = getattr(settings, 'CONSTANCE_BACKEND', 'constance.backends.redisd.RedisBackend')

CONFIG = parse_config(getattr(settings, 'CONSTANCE_CONFIG', {}))

CONNECTION_CLASS = getattr(settings, 'CONSTANCE_REDIS_CONNECTION_CLASS', getattr(settings, 'CONSTANCE_CONNECTION_CLASS', None))

REDIS_CONNECTION = getattr(settings, 'CONSTANCE_REDIS_CONNECTION', getattr(settings, 'CONSTANCE_CONNECTION', {}))

DATABASE_CACHE_BACKEND = getattr(settings, 'CONSTANCE_DATABASE_CACHE_BACKEND', None)

DATABASE_PREFIX = getattr(settings, 'CONSTANCE_DATABASE_PREFIX', '')

SUPERUSER_ONLY = getattr(settings, 'CONSTANCE_SUPERUSER_ONLY', getattr(settings, 'CONSTANCE_ACCESS_SUPERUSER_ONLY', True))

CACHE_TIMEOUT = getattr(settings, 'CONSTANCE_CACHE_TIMEOUT', 1)

CONFIG_CLASS = getattr(settings, 'CONSTANCE_CONFIG_CLASS', 'constance.config.Config')
